use Mail::SpamAssassin;
use DBI;

my $spamtest;

sub getStatusLine {
	my ($status) = @_;
	my @test_names = sort(@{$status->{test_names_hit}});

	my $line = ($status->is_spam() ? "Yes, " : "No, ");
	$line .= $status->{score} . "," . $status->{conf}->{required_score};

	foreach $_ (@test_names) {
		$line .= "," . $_ . ": " . $status->{conf}->{scores}->{$_};
	}

	$line .= ",autolearn=" . $status->{auto_learn_status};
	
	return $line;
}


sub initEngine
{
	my ($sa_dir, $db_dir, $enableSurbl, $enableDebug) = @_;
	my $bayes_db = "$db_dir/bayes";
        my $rules_dir = "$sa_dir/rules";
        my $prefix_dir = "$sa_dir/site";
	my $rv = 1;
	
	$spamtest = new Mail::SpamAssassin ({
	    rules_filename    => $rules_dir,
	    site_rules_filename => $rules_dir,
	    userstate_dir     => $bayes_db,
	    local_tests_only  => !($enableSurbl),
	    debug             => $enableDebug,
	    dont_copy_prefs   => 1,
	    PREFIX            => $prefix_dir,
	    DEF_RULES_DIR     => $rules_dir,
	    LOCAL_RULES_DIR   => $rules_dir
	});

	$spamtest->{conf}->{bayes_store_module} = "Mail::SpamAssassin::BayesStore::SDBM";
	$spamtest->{conf}->{auto_whitelist_factory} = "Mail::SpamAssassin::DBBasedAddrList";

	$spamtest->compile_now(0);
	$spamtest->init_learner();

	return $rv;
	
}

sub checkMessage
{
	my ($spam_file_name) = @_;

  	my $spam_file;
	open($spam_file, $spam_file_name);

	my $mail = $spamtest->parse($spam_file);

	close($spam_file);

	my $status = $spamtest->check($mail);
	my $line = getStatusLine($status);

	$status->finish();
	$mail->finish();

	return $line; 
};

sub learnMessage
{
	my ($spam_file_name, $isspam) = @_;	
	
	my $spam_file;
	open($spam_file, $spam_file_name);

	my $mail = $spamtest->parse($spam_file);
	
	close($spam_file);

	my $status = $spamtest->learn($mail, undef, $isspam, 0);
	
	$status->finish();
	$mail->finish();
}


sub getBayesStatistics
{
	$spamtest->{bayes_scanner}->{store}->tie_db_readonly();
	my @vars = $spamtest->{bayes_scanner}->{store}->get_storage_variables();
	$spamtest->{bayes_scanner}->{store}->untie_db();
	return @vars;
	
}

sub clearExpiredEntries
{
	$spamtest->rebuild_learner_caches();
	$spamtest->finish_learner();
}

use strict;
use Fcntl;
use Time::Local;
use File::Glob ':glob';	# handle correctly space character in directory path
use SDBM_File;
use Mail::SpamAssassin::Util;
use Mail::SpamAssassin::Logger;

sub string_ends_with
{
    my ($complete_string, $ending_token) = @_;

	my $len = length($ending_token);
	my $string_end = substr($complete_string, -$len);

	return $string_end eq $ending_token;
}

sub rename_db_file
{
	my ($source_name, $destination_name) = @_;

	# Now rename the new one.  Try several extensions.
	for my $ext ('.pag', '.dir')
	{
		my $rename_attempt_count = 0;
		while (!rename($source_name.$ext, $destination_name.$ext))
		{
			if ($rename_attempt_count > 500) {
				dbg("auto-whitelist: Cannot rename file from " . $source_name.$ext . " to " .
					$destination_name.$ext . ". $!.\n");
				return 0;
			 }
			 
			$rename_attempt_count ++;
		}
	}

	return 1;
}

sub clearAWLEntries
{
	# Timeout after which the procedure copying the original --> new entries is stopped.
	# Complete run of this function will take a little bit more time.
	use constant AWL_ROUGH_TIMEOUT => 4 * 60;

	# Print a lot of debug information, for development purposes.
	use constant DEVEL_DEBUG => 0;

	# Use the following autowhitelist database file.
	my $db = $spamtest->{userstate_dir} . "/auto-whitelist";
 
  # Move original AWL database to a new file. If the file is damaged, it will not
  # crash Perl interpreter next time.
  my $db_source = $db.'_toks.'."source";

	# clean out any leftover source db copies from previous runs
	my @prevsources = glob($db_source . "*");
	foreach my $prevsourcefile (@prevsources)
	{
		unlink $prevsourcefile;
	}

	# use a source PID-based suffix just in case another one was
	# created previously by an interrupted expire
	my $db_source_name = $db_source . "$$";
	 
	if (! rename_db_file($db, $db_source_name) )
	{
		return;
	}

	# Hash table which is associated with db file.
	my %h;
	tie %h, "SDBM_File", $db_source_name, O_RDONLY, (0700 & 0666);
	if (not tied %h)
	{
		warn "auto-whitelist: cannot open original database file $db_source_name\n";
		return;
	}
	
	# Removal of entries does not decrease the size of the database file on the disk.
	# --> Create new AWL database and remove the old one to shrink the database file.
	#
	my $tmpdb = $db.'_toks.'."expire";

	# clean out any leftover db copies from previous runs
	my @prevtmps = glob($tmpdb . "*");
	foreach my $prevtmpfile (@prevtmps) 
	{
		unlink $prevtmpfile;
	}

	# use a temporary PID-based suffix just in case another one was
	# created previously by an interrupted expire
	my $tmpdbname = $tmpdb . "$$";

	# create/write new db file
	my %new_toks;
	tie %new_toks, "SDBM_File", $tmpdbname, O_RDWR|O_CREAT|O_EXCL,
			  (0700 & 0666);
	if (not tied %new_toks)
	{
		warn "auto-whitelist: cannot open new database file $tmpdbname (failed)\n";
		return;
	}
	
	# entries with kerioatime older than the constant are expired (removed = not copied).
	use constant MONTH_IN_SECONDS => 30 * 24 * 60 * 60;
	my $ageout = time - MONTH_IN_SECONDS;	# all items older than ageout are removed
	my $copied = 0;	# copied items counter
	my $removed = 0; # removed items counter

	# associative array containing entries from original db without kerioatime
	my %upgrade; 

	# utility variables for function call timeout
	my $upgrade_skipsome = 0;
	my $totcount = 0;
	my $invalidcount = 0;
	my $startchecktime = time();

	my $upgradecount = 0;

	dbg("auto-whitelist: going to AWL expiry");

	# Go ahead and copy unexpired entries to new temporary db ...
	while((my $key, my $value) = each(%h)) 
	{
		# Finish the loop when time is out.
		if (($totcount % 1000) == 0) {
			if (DEVEL_DEBUG == 1)
			{
				dbg("auto-whitelist: already processed number of records: $totcount");
			}
			
			my $time_out = AWL_ROUGH_TIMEOUT;
			
			# When there are some records for upgrade then reserve
			# some time for upgrade hash sorting.
			if ($upgradecount > 50000)
			{
				$time_out = $time_out / 2;
			}
			
			if ((time() - $startchecktime) > $time_out)
			{
				dbg("auto-whitelist: too large number of records in database: " . 
					"some were skipped, stopped after $totcount records\n");
				last;
			}
		}

		# Each AWL record is stored in 3 hash slots.
		# Now we want to identify the basic slot email@domain|ip=number and ignore
		# email@domain|ip=number|totscore and email@domain|ip=number|kerioatime.
		
		my $totalscore_found = string_ends_with( $key, "|totscore" );
		my $kerioatime_found = string_ends_with( $key, "|kerioatime" );
	
		if ((not $totalscore_found) and (not $kerioatime_found))
		{
			$totcount ++;
	
			# data from hash table
			my $count = $h{$key};
			my $totscore = $h{"$key|totscore"};
			my $atime = $h{"$key|kerioatime"};

			# print info message
			if (DEVEL_DEBUG == 1) 
			{ 
				my $debug = "auto-whitelist: hash key: $key, count: $count, total score: $totscore, access time: ";
				if  ( defined($atime) )
				{
					$debug .= $atime;
				}		
				else
				{
					$debug .= "not defined (version 0 record)";
				}
				
				dbg($debug); 
			}
			
			if ( (not defined($count)) or (not defined($totscore)) )
			{
				if (DEVEL_DEBUG == 1)
				{
					dbg("auto-whitelist: missing count or total score: record skipped");
				}
				$invalidcount ++;
				next;
			}

			if  ( defined($atime) )
			{
				# Copy new records -- records newer than atime.
				if ($atime > $ageout)
				{
					$new_toks{ $key } = $count;
					$new_toks{ $key . "|totscore" } = $totscore;
					$new_toks{ $key . "|kerioatime" } = $atime;	
					$copied ++;
				}
				else
				{
					$removed ++;
				}				
			}
			else
			{
				# Records for upgrade.
				if ($upgradecount < 1)
				{
					dbg("auto-whitelist: some records are suitable for upgrade");
				}
				
				# Entry $key is without kerioatime. 
				# Use multiple values for the same hash key.
				push( @{$upgrade{$count}}, ($key, $totscore) );
				$upgradecount ++;
			}
		}
		else
		{
			# skip this hash slot
		}
	}
	
	dbg("auto-whitelist: expiration check finished in " . (time() - $startchecktime) . 
		" seconds from start, $copied records copied, $removed records expired, $invalidcount invalid records\n");

	# Upgrade entries from original SA AWL database.
	my $tmpcopied = $copied;

	if ($upgradecount > 0)
	{
		dbg("auto-whitelist: going to expiration upgrade for $upgradecount " . 
			"records\n");

		$totcount = 0;
		my $startupgradetime = time();

		foreach my $count (sort { $b <=> $a } keys %upgrade) {

			# Finish the loop when time is out.
			$totcount ++;
			if (($totcount % 1000) == 0) {
				if (DEVEL_DEBUG == 1)
				{
					dbg("auto-whitelist: already processed number of records: $totcount");
				}		
				if ((time() - $startchecktime) > AWL_ROUGH_TIMEOUT)
				{
					dbg("auto-whitelist: too large number of records in database: " . 
						"some were skipped, stopped after $totcount iterations\n");
					last;
				}
			}	

			# Convert old entry to new AWL db entry.
			my $key = "";
			foreach my $entry (@{$upgrade{$count}}) 
			{
				if ((length($key)) == 0)
				{
					$key = $entry;
					if (DEVEL_DEBUG == 1) { print STDERR "key: " . $key . ": "; }
				}
				else
				{
					$new_toks{ $key } = $count;
					$new_toks{ $key . "|totscore" } = $entry;
					$new_toks{ $key . "|kerioatime" } = time;
					$copied ++;
					
					if (DEVEL_DEBUG == 1) { print STDERR "count: $count, totscore: $entry\n"; }
				
					$key = "";	# reset the value
				}
			}
		}

		dbg("auto-whitelist: expiration upgrade finished in " . (time() - $startupgradetime) . 
			" seconds from start, " . ($copied - $tmpcopied) . " additional records copied\n");
	}

	# Close original AWL database and new (tmp) database files.
	untie %h;
	untie %new_toks;

	# Dispose at least some memory explicitly.
	undef (%h);
	undef (%new_toks);
	undef (%upgrade);

	if ($copied < 10000)
	{
		dbg("auto-whitelist: only $copied records found in the AWL database: it's too little; abort expiration procedure; " .
			"total time: " . (time() - $startchecktime) . " seconds\n");

		@prevtmps = glob($tmpdb . "*");
		foreach my $prevtmpfile (@prevtmps) 
		{
			unlink $prevtmpfile;
		}
  
		# Move back source database.
		rename_db_file($db_source_name, $db);

		return;
	}
	else
	{
		# Delete source database.
		for my $ext ('.pag', '.dir')
		{
			my $oldf =  $db_source_name.$ext;

			my $delete_attempt_count = 0;
			while (!unlink($oldf))
			{
				if ($delete_attempt_count > 500) {
					dbg("auto-whitelist: cannot delete " . $oldf ." file.\n");
					last;
				}
				
				$delete_attempt_count ++;
			}
		}
	}
	
	dbg("auto-whitelist: total number of records in new database after expiration run: " 
		. $copied . ", total time: " . (time() - $startchecktime) . " seconds\n");

	# Now swap the temporary and original AWL files.
	dbg("auto-whitelist: switch temporary and original files...\n");

	# Don't allow moving files around to be interrupted.  Scope the signal changes.
	my $success = 1;
	{
		local $SIG{'INT'} = 'IGNORE';
		local $SIG{'TERM'} = 'IGNORE';
		local $SIG{'HUP'} = 'IGNORE' if (!Mail::SpamAssassin::Util::am_running_on_windows());
		
		# Now rename the new one.  Try several extensions.
		for my $ext ('.pag', '.dir') 
		{
			my $newf = $tmpdbname.$ext;
			my $oldf = $db.$ext;
			next unless (-f $newf);

			# When file is accessed by the other SpamAssassin process,
			# the AWL db file may be in use. Try to delete it multiple times.
			my $delete_attempt_count = 0;
			while (!unlink($oldf))
			{
				if ($delete_attempt_count > 500) {
					last;
				}
				$delete_attempt_count ++;
			}
		
			if (!rename ($newf, $oldf)) 
			{
				dbg("auto-whitelist: rename $newf to $oldf failed: $!");
				$success = 0;
			}
		}
	}

	if ( $success == 1 )
	{
		dbg("auto-whitelist: files swapped, expiration successful\n");
	}
	else
	{
		dbg("auto-whitelist: expiration run failed\n");
	}
}
